import torch

def equilibrate(
        A,
        alpha_squared=1,
        beta_squared=1,
        gamma=1.0,
        M=1000.,
        iterations=20,
):
    """
        Algorithm taken from:

        S. Diamond, S. Boyd. Stochastic Matrix-Free Equilibration,
            *Journal of Optimization Theory and Applications*,
            172(2), 436-454, 2017.

        All constants are random.
    """
    n, m = A.shape
    u_k = torch.zeros(n)
    v_k = torch.zeros(m)
    u_bar = torch.zeros(n)
    v_bar = torch.zeros(m)

    for t in range(1, iterations):
        D = torch.exp(u_k)
        E = torch.exp(v_k)
        s = torch.randint(-1, 2, size=n)
        w = torch.randint(-1, 2, size=m)

        u: torch.Tensor = u_k - 2 * (
                        torch.abs(D * (A @ (E * s)))**2 - alpha_squared + gamma * u_k
                    ) / (gamma * (t + 1))
        u_k = torch.clip(u, -M, M)

        # Note A = A.T
        v: torch.Tensor = v_k - 2 * (
                        torch.abs(E * (A @ (D * w)))**2 - beta_squared + gamma * v_k
                    ) / (gamma * (t + 1))
        v_k = torch.clip(v, -M, M)
        u_bar = (2 / (t + 2)) * u_k + (t / (t+2)) * u_bar
        v_bar = (2 / (t + 2)) * v_k + (t / (t+2)) * v_bar
            
    D = torch.exp(u_bar)
    E = torch.exp(v_bar)
    return D, E

def symmetric_equilibrate(
    A, 
    alpha = 1,
    gamma = 0.4,
    M = 15,
    iterations=200,
    device=None
):
    n, m = A.shape
    assert n == m
    u_k = torch.zeros(n, device=device)
    u_bar = torch.zeros(n, device=device)
    alpha_squared = alpha**2

    for t in range(1, iterations):
        D = torch.exp(u_k)
        s = 2 * torch.randint(2, (n,), device=device) - 1
        u = u_k - 2 * (
                        torch.abs(D * (A @ (D * s)))**2 - alpha_squared + gamma * u_k
                    ) / (gamma * (t + 1))
        u_k = torch.clip(u, -M, M)
        u_bar = (2 / (t + 2)) * u_k + (t / (t+2)) * u_bar

    return torch.exp(u_bar)



from linops.linops import (
        operator_matrix_product,
        aslinearoperator,
        LinearOperator,
        MatrixOperator,
        SelectionOperator,
        VectorJacobianOperator,
)

from linops.linop_impls import (
        IdentityOperator,
        DiagonalOperator,
        KKTOperator,
        ZeroOperator,
)
from linops.blocks import (
    vstack,
    hstack
)
import linops.equilibration



import torch
import linops

def xdiag(A, m: int):
    n, n1 = A.shape
    assert n == n1
    m = m // 2
    def normalize_columns(M):
        return M / torch.linalg.vector_norm(M, dim=0)

    def diag_of_AB(A, B):
        return torch.sum(A * B, dim=0)

    Omega = (2 * torch.randint(2, size=(n, m), device=A.device) - 1).float()
    Y = linops.operator_matrix_product(A, Omega)
    Q, R = torch.linalg.qr(Y)

    Z = linops.operator_matrix_product(A.T, Q)
    T = Z.T @ Omega
    S = normalize_columns(torch.linalg.inv(R).T)
    dQZ = diag_of_AB(Q.T, Z.T)
    dQSSZ = diag_of_AB((Q @ S).T, (Z @ S).T)
    dOmegaQT = diag_of_AB(Omega.T, (Q @ T).T)
    dOmegaY = diag_of_AB(Omega.T, Y.T)
    dOmegaQSST = diag_of_AB(Omega.T, 
                            (Q @ S @ torch.diag(diag_of_AB(S, T))).T)

    return dQZ + (-dQSSZ + dOmegaY -dOmegaQT +dOmegaQSST) / m



import torch
from linops import LinearOperator

class ZeroOperator(LinearOperator):
    supports_operator_matrix = True
    def __init__(self, shape, adjoint=None):
        self._shape = shape
        self._adjoint = ZeroOperator((shape[1], shape[0]), self) \
                if adjoint is None else adjoint

    def _matmul_impl(self, v):
        shape = (self.shape[0], *v.shape[1:])
        return torch.zeros(shape, device=v.device)

class IdentityOperator(LinearOperator):
    supports_operator_matrix = True
    efficient_inverse = True
    def __init__(self, n):
        self._adjoint = self
        self._shape = (n, n)

    def _matmul_impl(self, v):
        return v

    def solve_A_x_eq_b(self, b, x0=None):
        return b


class DiagonalOperator(LinearOperator):
    supports_operator_matrix = True
    efficient_inverse = True
    def __init__(self, diag):
        self._adjoint = self
        self._diag = diag
        m, = diag.shape
        self._shape = (m, m)

    def _matmul_impl(self, v):
        if len(v.shape) == 1:
            return self._diag * v
        else:
            return self._diag[:, None] * v

    def solve_A_x_eq_b(self, b, x0=None):
        if len(b.shape) == 1:
            return b / self._diag
        else:
            return b / self._diag[:, None]

class KKTOperator(LinearOperator):
    def __init__(self, H: LinearOperator, A: LinearOperator):
        k, ell = H.shape
        assert k == ell
        m, n = A.shape
        assert n == k
        self._shape = (m + n, m + n)
        self._adjoint = self
        self._A = A
        self._H = H
        self._m = m
        self._n = n

    def _matmul_impl(self, y):
        n = self._n
        return torch.hstack([
            self._H @ y[:n] - self._A.T @ y[n:],
            -self._A @ y[:n]
        ])



""" 
Code based on scipy.sparse.linalg.lsqr

The original Fortran code was written by C. C. Paige and M. A. Saunders as
described in
C. C. Paige and M. A. Saunders, LSQR: An algorithm for sparse linear
equations and sparse least squares, TOMS 8(1), 43--71 (1982).
C. C. Paige and M. A. Saunders, Algorithm 583; LSQR: Sparse linear
equations and least-squares problems, TOMS 8(2), 195--209 (1982).
It is licensed under the following BSD license:
Copyright (c) 2006, Systems Optimization Laboratory

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

3   Neither the name of Stanford University nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
The Fortran code was translated to Python for use in CVXOPT by Jeffery
Kline with contributions by Mridul Aanjaneya and Bob Myhill.
Adapted for SciPy by Stefan van der Walt.
"""

import torch

eps = torch.finfo(torch.float64).eps

def _sym_ortho(a, b):
    if b == 0:
        return torch.sign(a), 0, abs(a)
    elif a == 0:
        return 0, torch.sign(b), abs(b)
    elif abs(b) > abs(a):
        tau = a / b
        s = torch.sign(b) / torch.sqrt(1 + tau * tau)
        c = s * tau
        r = b / s
    else:
        tau = b / a 
        c = torch.sign(a) / torch.sqrt(1+tau*tau)
        s = c * tau
        r = a / c
    return c, s, r

def lsqr(A, b, damp=0.0, atol=1e-6, btol=1e-6, conlim=1e8,
         iter_lim=None, calc_var=False, x0=None):

        m, n = A.shape
        if iter_lim is None:
            iter_lim = 2 * n
        var = torch.zeros(n, device=b.device, dtype=b.dtype)

        itn = 0
        istop = 0
        ctol = 0
        if conlim > 0:
            ctol = 1/conlim
        anorm = 0
        acond = 0
        dampsq = damp**2
        ddnorm = 0
        res2 = 0
        xnorm = 0
        xxnorm = 0
        z = 0
        cs2 = -1
        sn2 = 0
        u = b
        bnorm = torch.linalg.norm(b)

        if x0 is None:
            x = torch.zeros(n, device=b.device, dtype=b.dtype)
            beta = bnorm.clone()
        else:
            x = x0
            u = u - A@x
            beta = torch.linalg.norm(u)

        if beta > 0:
            u = (1/beta) * u
            v = A.T@u
            alfa = torch.linalg.norm(v)
        else:
            v = x.clone()
            alfa = 0
        
        if alfa > 0:
            v = (1/alfa) * v
        w = v.clone()

        rhobar = alfa
        phibar = beta
        rnorm = beta
        r1norm = rnorm
        r2norm = rnorm

        arnorm = alfa * beta
        if arnorm == 0:
            return x

        # Main iteration loop
        while itn < iter_lim:
            itn = itn + 1
            u = A@v - alfa * u
            beta = torch.linalg.norm(u)

            if beta > 0:
                u = (1/beta)*u
                anorm = torch.sqrt(anorm**2 + alfa**2 + beta**2 + dampsq)
                v = A.T@u - beta * v
                alfa = torch.linalg.norm(v)
                if alfa > 0:
                    v = (1 / alfa * v)

            if damp > 0:
                rhobar1 = torch.sqrt(rhobar**2 + dampsq)
                cs1 = rhobar / rhobar1
                sn1 = damp / rhobar1
                psi = sn1 * phibar
                phibar = cs1 * phibar
            else:
                rhobar1 = rhobar
                psi = 0
            cs, sn, rho = _sym_ortho(rhobar1, beta)

            theta = sn * alfa
            rhobar = -cs * alfa
            phi = cs * phibar
            phibar = sn * phibar
            tau = sn * phi

            t1 = phi / rho
            t2 = -theta / rho
            dk = (1/rho) * w

            x = x + t1 * w
            w = v + t2 * w
            ddnorm = ddnorm + torch.linalg.norm(dk)**2

            if calc_var:
                var = var + dk**2

            delta = sn2 * rho
            gambar = -cs2 * rho
            rhs = phi - delta * z
            zbar = rhs / gambar
            xnorm = torch.sqrt(xxnorm + zbar**2)
            gamma = torch.sqrt(gambar**2 +theta**2)
            cs2 = gambar / gamma
            sn2 = theta / gamma
            z = rhs / gamma
            xxnorm = xxnorm + z**2

            acond = anorm * torch.sqrt(ddnorm)
            res1 = phibar**2
            res2 = res2 + psi**2
            rnorm = torch.sqrt(res1 + res2)
            arnorm = alfa * torch.abs(tau)

            if damp > 0:
                r1sq = rnorm**2 - dampsq * xxnorm
                r1norm = torch.sqrt(torch.abs(r1sq))
                if r1sq < 0:
                    r1norm = -r1norm
            else:
                r1norm = rnorm
            r2norm = rnorm

            test1 = rnorm / bnorm
            test2 = arnorm / (anorm * rnorm + eps)
            test3 = 1 / (acond + eps)
            t1 = test1 / (1 + anorm * xnorm/bnorm)
            rtol = btol + atol * anorm / bnorm

            if itn >= iter_lim:
                istop = 7
            if 1 + test3 <= 1:
                istop = 6
            if 1 + test2 <= 1:
                istop = 5
            if 1 + t1 <= 1:
                istop = 4

            if test3 <= ctol:
                istop = 3
            if test2 <= atol:
                istop = 2
            if test1 <= rtol:
                istop = 1

            if istop != 0:
                break
    
        return x



import torch
import operator
import linops as lo
from linops.cg import CG
from linops.minres import minres
import linops.nystrom_precondition as nystrom 



def operator_matrix_product(A, M):
    assert A.shape[1] == M.shape[0]
    if A.supports_operator_matrix:
        return A @ M
    out = torch.empty((A.shape[0], M.shape[1]), device=M.device, dtype=M.dtype)
    for i in range(M.shape[1]):
        out[:, i] = A @ M[:, i]
    return out


def aslinearoperator(A):
    if isinstance(A, LinearOperator):
        return A
    else:
        return MatrixOperator(A)

class LinearOperator:
    _adjoint = None
    _shape: tuple[int, int] = None
    _nystrom_sketch = None
    _last_solve_of_x = None
    device = None
    supports_operator_matrix = False
    efficient_inverse = False

    def __call__(self, x):
        return self @ x

    def __matmul__(self, b):
        if isinstance(b, LinearOperator):
            return _JoinOperator(self, b)
        elif len(b.shape) == 1:
            return self._matmul_impl(b)
        elif len(b.shape) == 2:
            assert self.shape[1] == b.shape[0]
            if self.supports_operator_matrix:
                return self._matmul_impl(b)
            else:
                out = torch.empty((self.shape[0], b.shape[1]), device=b.device, dtype=b.dtype)
                for i in range(b.shape[1]):
                    out[:, i] = self._matmul_impl(b[:, i])
                return out

    def __rmatmul__(self, a):
        if isinstance(a, LinearOperator):
            return _JoinOperator(a, self)
        else:
            return (self.T @ a.T).T

    def __mul__(self, c):
        return _ScaleOperator(c, self.shape) @ self
    
    def __truediv__(self, c):
        return _ScaledownOperator(c, self.shape) @ self

    def __rmul__(self, c):
        return self @ _ScaleOperator(c, self.shape)

    # __radd__ not needed because we only support sums of linear operators
    def __add__(self, b):
        if isinstance(b, LinearOperator):
            return _BinaryOperator(self, b, operator.add)
        else:
            return NotImplemented

    # __rsub__ not needed because we only support sums of linear operators
    def __sub__(self, b):
        if isinstance(b, LinearOperator):
            return _BinaryOperator(self, b, operator.sub)
        else:
            return NotImplemented

    def __neg__(self):
        return _UrnaryOperator(self, operator.neg)

    def __pos__(self):
        return _UrnaryOperator(self, operator.pos)

    def __pow__(self, n):
        return _PowOperator(self, n)

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            # Only recieved one index, so we're only applying to the output
            S = _generate_selector(self.shape[0], key[0])
            return S @ self
        leftS = _generate_selector(self.shape[0], key[0])
        rightS = _generate_selector(self.shape[1], key[1], flip=True)
        return leftS @ self @ rightS.T

    @property
    def shape(self) -> tuple[int, int]:
        return self._shape

    @property
    def T(self):
        """
        Can probably be auto-generated with
            D_x(v @ A @ x) = A^T v
        """
        if self._adjoint is None:
            x = torch.ones(self.shape[1], device=self.device, requires_grad=True)
            g = self @ x
            self._adjoint = VectorJacobianOperator(g, x, self)
        return self._adjoint

    def _matmul_impl(self, v):
        raise NotImplementedError()

    def solve_I_p_lambda_AT_A_x_eq_b(self, lambda_, b, x0=None, **kwargs):
        precondition = None
        if 'precondition' in kwargs and kwargs['precondition'] is not None:
            assert kwargs['precondition'] == 'nystrom'
            if self._nystrom_sketch is None:
                self._nystrom_sketch = nystrom.construct_approximation(
                    self,
                    l_0=kwargs.get('l_0', 10),
                    l_max=kwargs.get('l_max', 500),
                    power_iter_count=kwargs.get('power_iter_count', 50),
                    error_tol=kwargs.get('error_tol', 1.0),
                    )
            precondition = nystrom.NystromPreconditioner(
                    self._nystrom_sketch, lambda_)
        self._last_solve_of_x = CG(
                lo.IdentityOperator(self.shape[1]) + lambda_ * (self.T @ self),
                precondition)(b)
        return self._last_solve_of_x

    def solve_A_x_eq_b(self, b, x0=None):
        if self is self._adjoint:
            return minres(self, b, x0=x0)
        else:
            raise NotImplementedError()

class _ScaleOperator(LinearOperator):
    supports_operator_matrix = True
    efficient_inverse = True
    def __init__(self, c, shape):
        self._c = c
        self._adjoint = self
        self._shape = shape

    def _matmul_impl(self, v):
        return self._c * v

    def solve_A_x_eq_b(self, b, x0=None):
        return b / self._c

class _ScaledownOperator(LinearOperator):
    supports_operator_matrix = True
    efficient_inverse = True
    def __init__(self, c, shape):
        self._c = c
        self._adjoint = self
        self._shape = shape

    def _matmul_impl(self, v):
        return v / self._c

    def solve_A_x_eq_b(self, b, x0=None):
        return b * self._c

class MatrixOperator(LinearOperator):
    supports_operator_matrix = True
    def __init__(self, matrix, adjoint=None):
        self._matrix = matrix
        assert len(matrix.shape) == 2
        self._shape = matrix.shape
        if adjoint is None:
            self._adjoint = MatrixOperator(self._matrix.T, self)
        else:
            self._adjoint = adjoint
        self.__ATA_matrix = None
        self.__lambda_cache = None
        self.__IplATA_matrix = None
        self.__diagonal_ATA = None

    def _matmul_impl(self, v):
        return self._matrix @ v

    def _direct_solve_I_p_lambda_AT_A_x_eq_b(self, lambda_, b, **kwargs):
        if self.__ATA_matrix is None:
            A = self._matrix.clone()
            self.__ATA_matrix = A.T @ A

        if self.__lambda_cache != lambda_:
            lATA = lambda_ * (self.__ATA_matrix)
            torch.diagonal(lATA)[:] += 1
            self.__IplATA_matrix = lATA
        return torch.linalg.solve(self.__IplATA_matrix, b)

    def solve_A_x_eq_b(self, b, x0=None):
        return torch.linalg.solve(self._matrix, b)

    def solve_I_p_lambda_AT_A_x_eq_b(self, lambda_, b, x0=None, **kwargs):
        return self._direct_solve_I_p_lambda_AT_A_x_eq_b(lambda_, b,
                                                         x0=x0, **kwargs)

    def _cg_solve_I_p_lambda_AT_A_x_eq_b(self, lambda_, b, x0=None, **kwargs):
        if self.__diagonal_ATA is None:
            self.__diagonal_ATA = torch.linalg.vector_norm(self._matrix, axis=0)
        M = DiagonalOperator(lambda_ * self.__diagonal_ATA + 1)
        self._last_solve_of_x = CG(
                IdentityOperator(self.shape[1]) + lambda_ * (self.T @ self),
                M)(b)
        return self._last_solve_of_x

class _PowOperator(LinearOperator):
    def __init__(self, linop, k:int):
        self.supports_operator_matrix = linop.supports_operator_matrix
        self._linop = linop
        self._k = k
        assert k >= 0
        assert linop.shape[0] == linop.shape[1]
        self._adjoint = _PowOperator(self._linop.T, self._k)
        self._shape = linop.shape
        self.efficient_inverse = linop.efficient_inverse

    def _matmul_impl(self, v):
        k = self._k
        while k > 0:
            v = self._linop @ v
            k -= 1
        return v

    def solve_A_x_eq_b(self, b, x0=None):
        if self.efficient_inverse:
            k = self._k
            while k > 0:
                b = self._linop.solve_A_x_eq_b(b, x0)
                k -= 1
            return b
        else:
            return super().solve_A_x_eq_b(b, x0=x0)

class _BinaryOperator(LinearOperator):
    """transpose must distribute over op"""
    def __init__(self, left, right, op, adjoint=None):
        assert left.shape == right.shape
        self.supports_operator_matrix = \
                left.supports_operator_matrix and right.supports_operator_matrix
        self._left = left
        self._right = right
        self._op = op
        if adjoint is None:
            self._adjoint = _BinaryOperator(
                    self._left.T, self._right.T, self._op, self)
        self._shape = left.shape

    def _matmul_impl(self, y):
        return self._op(self._left @ y, self._right @ y)
    
class _UrnaryOperator(LinearOperator):
    """ transpose must pass through op """
    def __init__(self, linop, op, adjoint=None):
        self._linop = linop
        self.supports_operator_matrix = linop.supports_operator_matrix
        self._op = op
        if adjoint is None:
            self._adjoint = _UrnaryOperator(self._linop.T, self._op, self)
        else:
            self._adjoint = adjoint
        self._shape = linop.shape

    def _matmul_impl(self, y):
        return self._op(self._linop @ y)
    
class _JoinOperator(LinearOperator):
    def __init__(self, left, right, adjoint=None):
        self._left = left
        self._right = right
        self.efficient_inverse = \
                left.efficient_inverse and right.efficient_inverse

        self.supports_operator_matrix = \
                left.supports_operator_matrix and right.supports_operator_matrix
        if adjoint is None:
            self._adjoint = _JoinOperator(self._right.T, self._left.T, self)
        else:
            self._adjoint = adjoint
        self._shape = (left.shape[0], right.shape[1])

    def _matmul_impl(self, y):
        return self._left @ (self._right @ y)

    def solve_A_x_eq_b(self, b, x0=None):
        if self.efficient_inverse:
            y = self._left.solve_A_x_eq_b(b, self._right @ x0)
            return self._right.solve_A_x_eq_b(y, x0)
        else:
            return super().solve_A_x_eq_b(b, x0=x0)

class VectorJacobianOperator(LinearOperator):
    def __init__(self, g, x, adjoint=None):
        self._shape = (x.shape[0], g.shape[0])
        self._adjoint = adjoint
        self.device = x.device
        self._g = g
        self._x = x

    def _matmul_impl(self, v):
        self._x.grad = None
        self._g.backward(gradient=v, retain_graph=True)
        out = self._x.grad
        self._x.grad = None
        return out

class SelectionOperator(LinearOperator):
    #supports_operator_matrix = True
    def __init__(self, shape, idxs):
        self._shape = shape
        self._adjoint = _AdjointSelectionOperator(idxs,
                (self._shape[1], self._shape[0]), self)
        self._idxs = idxs

    def _matmul_impl(self, X):
        return X[self._idxs]

    def solve_I_p_lambda_AT_A_x_eq_b(self, lambda_, b):
        LHS = torch.ones_like(b)
        LHS[self._idxs] += lambda_
        return b / LHS

class _AdjointSelectionOperator(LinearOperator):
    #supports_operator_matrix = True
    def __init__(self,  idxs, shape, adjoint):
        self._adjoint = adjoint
        self._idxs = idxs
        self._shape = shape

    def _matmul_impl(self, y):
        z = torch.zeros(self.shape[0], dtype=y.dtype, device=y.device)
        z[self._idxs] = y
        return z.reshape(-1)

def _generate_selector(size_of_vec, key, flip=False):
    shape_of_output = torch.empty(size_of_vec)[key].shape
    if len(shape_of_output) == 0:
        shape_of_output = 1
    else:
        shape_of_output = shape_of_output[0]
    return SelectionOperator(
            (size_of_vec, shape_of_output)
                if not flip else (shape_of_output, size_of_vec)
            , key)



from linops import LinearOperator
import linops as lo
import torch

class FusedOperator(LinearOperator):
    def __init__(self, shape: tuple[int, int],
                       in_slices: list[slice],
                       out_slices: list[slice],
                       ops: list[LinearOperator],
                       adjoint=None):
        self._shape = shape
        self._in_slices = in_slices
        self._out_slices = out_slices
        self._ops = ops
        if adjoint is None:
            self._adjoint = FusedOperator((shape[1], shape[0]), out_slices,
                    in_slices, ops, self)
        else:
            self._adjoint = adjoint
    
    def _matmul_impl(self, v):
        out = torch.zeros(self._shape[0], device=v.device, dtype=v.dtype)
        for in_s, out_s, op in zip(self._in_slices, self._out_slices, self._ops,
                # strict=True, Readd this when upgrading MPV to 3.10 
                ):
            out[out_s] += op @ v[in_s]
        return out

class ZeroOperator(LinearOperator):
    def __init__(self, shape: tuple[int, int], adjoint=None):
        self._shape = shape
        if adjoint is None:
            self._adjoint = ZeroOperator((shape[1], shape[0]), self)
        else:
            self._adjoint = adjoint
    
    def _matmul_impl(self, v):
        return torch.zeros(self._shape[0], device=v.device, dtype=v.dtype)


def vstack(tup):
    tup = [lo.aslinearoperator(A) for A in tup]
    left_shape = [A.shape[0] for A in tup] 
    right_shape = {A.shape[1] for A in tup}
    assert len(right_shape) == 1
    right_dim = next(iter(right_shape))
    left_dim = sum(left_shape)
    value = 0

    return FusedOperator(
        (left_dim, right_dim),
        [slice(None) for _ in tup],
        [slice(value, (value := value + size)) for size in left_shape],
        tup)

def hstack(tup):
    tup = [lo.aslinearoperator(A) for A in tup]
    left_shape = {A.shape[0] for A in tup} 
    right_shape = [A.shape[1] for A in tup]
    assert len(left_shape) == 1
    left_dim = next(iter(left_shape))
    right_dim = sum(right_shape)
    value = 0

    return FusedOperator(
        (left_dim, right_dim),
        [slice(value, (value := value + size)) for size in right_shape],
        [slice(None) for _ in tup],
        tup)

def bmat(blocks):
    """
    Optimizes out `ZeroOperator`s.
    """
    blocks = [[lo.aslinearoperator(A) for A in r] for r in blocks]

    raise NotImplementedError("bmat implementation is incomplete")



"""

Copyright 2021 Parth Nobel
Copyright 2019 Shane Barratt


Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

import torch
import time

def cg_batch_Kis1(A_mm, B, M_mm=None, X0=None, rtol=1e-3, atol=0.,
                  maxiter=None, verbose=False):
    """Solves a batch of PD matrix linear systems using the preconditioned CG algorithm.
    This function solves a batch of matrix linear systems of the form
        A X = B
    where A is a n x n positive definite matrix and B is a n x m matrix,
    and X is the n x m matrix representing the solution for the ith system.
    Args:
        A_mm: A callable that performs a batch matrix multiply of A and a n x m matrix.
        B: A n x m matrix representing the right hand sides.
        M_mm: (optional) A callable that performs a batch matrix multiply of the preconditioning
            matrices M and a n x m matrix. (default=identity matrix)
        X0: (optional) Initial guess for X, defaults to M_bmm(B). (default=None)
        rtol: (optional) Relative tolerance for norm of residual. (default=1e-3)
        atol: (optional) Absolute tolerance for norm of residual. (default=0)
        maxiter: (optional) Maximum number of iterations to perform. (default=5*n)
        verbose: (optional) Whether or not to print status messages. (default=False)
    """
    n, = B.shape

    if M_mm is None:
        M_mm = lambda x: x
    if X0 is None:
        X0 = M_mm(B)
    if maxiter is None:
        maxiter = 5 * n

    assert B.shape == (n,)
    assert X0.shape == (n,)
    assert rtol > 0 or atol > 0
    assert isinstance(maxiter, int)

    X_k = X0
    R_k = B - A_mm(X_k)
    Z_k = M_mm(R_k)

    P_k = torch.zeros_like(Z_k)

    P_k1 = P_k
    R_k1 = R_k
    R_k2 = R_k
    X_k1 = X0
    Z_k1 = Z_k
    Z_k2 = Z_k

    B_norm = torch.norm(B)
    epsilon = torch.max(rtol*B_norm, atol*torch.ones_like(B_norm))

    if verbose:
        print("%03s | %010s %06s" % ("it", "dist", "it/s"))

    optimal = False
    start = time.perf_counter()
    for k in range(1, maxiter + 1):
        start_iter = time.perf_counter()
        Z_k = M_mm(R_k)

        if k == 1:
            P_k = Z_k
            R_k1 = R_k
            X_k1 = X_k
            Z_k1 = Z_k
        else:
            R_k2 = R_k1
            Z_k2 = Z_k1
            P_k1 = P_k
            R_k1 = R_k
            Z_k1 = Z_k
            X_k1 = X_k
            denominator = (R_k2 * Z_k2).sum()
            denominator[denominator == 0] = 1e-8
            beta = (R_k1 * Z_k1).sum() / denominator
            P_k = Z_k1 + beta * P_k1

        denominator = (P_k * A_mm(P_k)).sum()
        denominator[denominator == 0] = 1e-8
        alpha = (R_k1 * Z_k1).sum() / denominator
        X_k = X_k1 + alpha * P_k
        R_k = R_k1 - alpha * A_mm(P_k)
        end_iter = time.perf_counter()

        residual_norm = torch.norm(A_mm(X_k) - B)

        if verbose:
            print("%03d | %8.4e %8.4e %4.2f" %
                  (k, residual_norm, epsilon,
                    1. / (end_iter - start_iter)))

        if (residual_norm <= epsilon).all():
            optimal = True
            break

    end = time.perf_counter()

    if verbose:
        if optimal:
            print("Terminated in %d steps (optimal). Took %.3f ms." %
                  (k, (end - start) * 1000))
        else:
            print("Terminated in %d steps (reached maxiter). Took %.3f ms." %
                  (k, (end - start) * 1000))


    if verbose:
        info = {
            "niter": k,
            "optimal": optimal,
            "residual": residual_norm
        }
        print(info)

    return X_k

def CG(A_mm, M_mm=None, rtol=1e-3, atol=0., maxiter=None, X0=None,
            verbose=False):
    bX0 = None
    class CG(torch.autograd.Function):
        
        @staticmethod
        def forward(_ctx, B):
            X = cg_batch_Kis1(A_mm, B, M_mm=M_mm, X0=X0, rtol=rtol,
                        atol=atol, maxiter=maxiter, verbose=verbose)
            return X

        @staticmethod
        def backward(_ctx, dX):
            nonlocal bX0
            dB = cg_batch_Kis1(A_mm, dX, M_mm=M_mm, X0=bX0, rtol=rtol,
                        atol=atol, maxiter=maxiter, verbose=verbose)
            bX0 = dB
            return dB
    return CG.apply



import torch
import linops as lo

def minres(A, b, M=None, x0=None, tol=1e-5, maxiters=None, verbose=True):
    """
    Code based on scipy.sparse.linalg.minres

    Copyright 2022 Parth Nobel
    Copyright 2001-2002 Enthought, Inc. 2003-2022, SciPy Developers.

    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:

    1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials provided
        with the distribution.

    3. Neither the name of the copyright holder nor the names of its
        contributors may be used to endorse or promote products derived
        from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    m, n = A.shape
    assert m == n
    if M is None:
        M = lo.IdentityOperator(m)
    if maxiters is None:
        maxiters = 5 * n
    if x0 is None:
        x = torch.zeros_like(b)
    else:
        x = x0
    iters = 0
    Anorm = 0
    Acond = 0
    eps = torch.tensor(torch.finfo(b.dtype).eps, device=x.device)

    r1 = b - A @ x
    y = M @ r1

    beta1 = r1 @ y
    assert beta1 >= 0
    if beta1 == 0:
        return x
    bnorm = torch.linalg.vector_norm(b)
    if bnorm == 0:
        return b

    beta1 = torch.sqrt(beta1)

    oldb = 0
    beta = beta1
    dbar = 0
    epsilon = 0
    phibar = beta1
    rhs1 = beta1
    rhs2 = 0
    tnorm2 = torch.tensor(0.0, device=x.device)
    gmax = torch.tensor(0.0, device=x.device)
    gmin = torch.tensor(torch.finfo(x.dtype).max, device=x.device)
    cs = -1
    sn = 0
    w = torch.zeros_like(b)
    w2 = torch.zeros_like(b)
    r2 = r1
    
    while iters < maxiters:
        iters += 1
        s = 1.0 / beta
        v = s * y
        y = A @ v
        if iters >= 2:
            y = y - (beta / oldb) * r1

        alpha = v @ y
        y = y - (alpha / beta) * r2
        r1 = r2
        r2 = y
        y = M @ r2
        oldb = beta
        beta = r2 @ y
        assert beta >= 0
        beta = torch.sqrt(beta)
        tnorm2 += alpha**2 + oldb**2 + beta**2
        if iters == 1:
            if beta / beta1 <= 10 * eps:
                #assert False, "I think this occurs when A = c * I"
                pass

        oldeps = epsilon
        delta = cs * dbar + sn * alpha
        gbar = sn * dbar - cs * alpha
        epsilon = sn * beta
        dbar = -cs * beta
        root = L2norm2entry(gbar, dbar)
        #Arnorm = phibar * root

        gamma = L2norm2entry(gbar, beta)
        gamma = torch.maximum(gamma, eps)
        cs = gbar / gamma
        sn = beta / gamma
        phi = cs * phibar
        phibar = sn * phibar


        denom = 1.0 / gamma
        w1 = w2
        w2 = w
        w = (v - oldeps * w1 - delta * w2) * denom
        x = x + phi * w

        gmax = torch.max(gmax, gamma)
        gmin = torch.min(gmin, gamma)
        z = rhs1 / gamma
        rhs1 = rhs2 - delta * z
        rhs2 = - epsilon * z

        Anorm = torch.sqrt(tnorm2)
        ynorm = torch.linalg.norm(x)
        epsa = Anorm * eps
        epsx = Anorm * ynorm * eps
        #epsr = Anorm * ynorm * tol
        diag = gbar
        if diag == 0:
            diag = epsa
        qrnorm = phibar
        rnorm = qrnorm
        if ynorm == 0 or Anorm == 0:
            test1 = torch.inf
        else:
            test1 = rnorm / (Anorm * ynorm)

        if Anorm == 0:
            test2 = torch.inf
        else:
            test2 = root / Anorm
        Acond = gmax / gmin
        t1 = 1 + test1
        t2 = 1 + test2
        if t2 <= 1:
            break
        if t1 <= 1:
            break

        if Acond >= 0.1 / eps:
            assert False, "System is ill-conditioned."
        if epsx >= beta1:
            assert False
        if test2 <= tol:
            break
        if test1 <= tol:
            break
    return x

def L2norm2entry(x, y):
    return torch.sqrt(x**2 + y**2)



from __future__ import annotations

import math

import torch
import linops as lo

"""
This module contains various ways to compute divergences.

Reference paper 1: https://arxiv.org/pdf/2010.09649.pdf
Reference paper 2: https://arxiv.org/pdf/2301.07825.pdf
"""

def _sample_dist(dist, shape, _gen_data_on_device, A_dev):
    if _gen_data_on_device is None:
        gen_dev = A_dev
        dest_dev = None
    else:
        gen_dev = _gen_data_on_device
        dest_dev = A_dev if A_dev is not None else torch.device('cpu')

    if dist == 'rademacher':
        return 2.0 * torch.randint(2, size=shape, device=gen_dev).to(dest_dev) - 1.0
    elif dist == 'gaussian' or dist == 'normal':
        return torch.randn(*shape, device=gen_dev).to(dest_dev)


def hutchinson(A: lo.LinearOperator, m: int=400, *, _gen_data_on_device=None):
    k, ell = A.shape
    assert k == ell
    if k <= m:
        return exact_trace(A)
    results = torch.empty(m, device=A.device)
    for i in range(m):
        z = _sample_dist('rademacher', (k,), _gen_data_on_device, A.device)
        results[i] = (z * (A @ z)).sum()
    return torch.mean(results), torch.std(results) / math.sqrt(m)



def hutchpp(A: lo.LinearOperator, m: int=102, *, _gen_data_on_device=None):
    n, ell = A.shape
    assert n == ell
    assert m % 3 == 0
    if n <= m:
        return exact_trace(A)
    k = m // 3
    S = _sample_dist('rademacher', (n, k), _gen_data_on_device, A.device)
    G = _sample_dist('rademacher', (n, k), _gen_data_on_device, A.device)

    BS = lo.operator_matrix_product(A, S)
    Q, _ = torch.linalg.qr(BS)
    G_prime = G - Q @ (Q.T @ G)
    exact = 0
    for i in range(Q.shape[1]):
        z = Q[:,i]
        exact = exact + (z * (A @ z)).sum()
    
    hutchinson = torch.empty(k, device=A.device)
    for i in range(k):
        z = G_prime[:,i]
        hutchinson[i] = (z * (A @ z)).sum()

    return exact + torch.mean(hutchinson), torch.std(hutchinson) / math.sqrt(m)

def exact_trace(A: lo.LinearOperator):
    m, n = A.shape
    assert m == n
    one = torch.ones(1, device=A.device)
    divergence = 0
    for i in range(m):
        divergence = divergence + A[i, i] @ one
    return divergence, 0.0

def xtrace(A, m: int=80, *, _gen_data_on_device=None):
    n, n1 = A.shape
    assert n == n1
    if n <= m:
        return exact_trace(A)

    m = m // 2
    def normalize_columns(M):
        return M / torch.linalg.vector_norm(M, dim=0)

    def diag_of_AB(A, B):
        return torch.sum(A * B, dim=0)

    Z = _sample_dist('gaussian', (n, m), _gen_data_on_device, A.device)
    Omega = math.sqrt(n) * normalize_columns(Z)
    Y = A @ Omega
    Q, R = torch.linalg.qr(Y)

    W = Q.T @ Omega
    S = normalize_columns(torch.linalg.inv(R).T)
    scale = (n - m + 1) / (n - torch.linalg.vector_norm(W, dim=0)**2 +
                           torch.abs(diag_of_AB(S, W) * torch.linalg.vector_norm(S, dim=0))**2)
    Z = A @ Q
    H = Q.T @ Z
    HW = H @ W
    T = Z.T @ Omega
    dSW = diag_of_AB(S, W)
    dSHS = diag_of_AB(S, H @ S)
    dTW = diag_of_AB(T, W)
    dWHW = diag_of_AB(W, HW)
    dSRmHW = diag_of_AB(S, R - HW)
    dTmHRS = diag_of_AB(T - H.T @ W, S)

    ests = torch.trace(H) - dSHS + (
            dWHW - dTW + dTmHRS * dSW + torch.abs(dSW)**2 * dSHS + dSW * dSRmHW) * scale

    return torch.mean(ests), torch.std(ests) / math.sqrt(m)


def xnystrace(A: lo.LinearOperator, m: int=80, *, _gen_data_on_device=None):
    n, n1 = A.shape
    assert n == n1
    if n <= m:
        return exact_trace(A)
    m = m // 2
    def normalize_columns(M):
        return M / torch.linalg.vector_norm(M, dim=0)

    def diag_of_AB(A, B):
        return torch.sum(A * B, dim=0)

    Z = _sample_dist('gaussian', (n, m), _gen_data_on_device, A.device)
    Omega = math.sqrt(n) * normalize_columns(Z)
    Y = A @ Omega

    nu = torch.finfo(Omega.dtype).eps * torch.linalg.vector_norm(Y) / math.sqrt(n)
    Y = Y + nu * Omega


    Q, R = torch.linalg.qr(Y)
    H = Omega.T @ Y
    C = torch.linalg.cholesky((H + H.T) / 2, upper=True)

    B = torch.linalg.solve_triangular(C, R, upper=True, left=False) # Double check this

    QQ, RR = torch.linalg.qr(Omega)
    WW = QQ.T @ Omega
    SS = normalize_columns(torch.linalg.inv(RR).T)

    scale = (n - m + 1) / (n - torch.linalg.vector_norm(WW, dim=0)**2 +
                           torch.abs(diag_of_AB(SS, WW) * torch.linalg.vector_norm(SS, dim=0))**2)
    W = Q.T @ Omega
    S =  torch.linalg.solve_triangular(C.T, B, upper=False, left=False) * torch.diag(torch.linalg.inv(H))**(-0.5) # Double check triangular solve
    dSW = diag_of_AB(S, W)
    ests = torch.linalg.norm(B, 'fro')**2 - torch.linalg.vector_norm(S, dim=0)**2 + torch.abs(dSW)**2 * scale - nu * n
    return torch.mean(ests), torch.std(ests) / math.sqrt(m)



from __future__ import annotations
import torch
import math
from dataclasses import dataclass
import linops as lo

@dataclass
class _NystromApproximation:
    U: torch.Tensor
    Lambda_hat: torch.Tensor

@dataclass
class NystromPreconditioner:
    A_hat: _NystromApproximation
    mu: torch.Tensor

    def __call__(self, v):
        U = self.A_hat.U
        Lambda_hat = self.A_hat.Lambda_hat
        mu = self.mu
        return (Lambda_hat[-1] + mu) * (U @ ((U.T @ v) / (Lambda_hat + mu))) + (v - U @ (U.T @ v))

def construct_approximation(
        A: lo.LinearOperator,
        l_0: int,
        l_max: int,
        power_iter_count: int,
        error_tol, dtype=None):
    n = A.shape[0]
    assert A.shape[0] == A.shape[1]
    device = A.device

    Y = torch.empty((n, 0), dtype=dtype, device=device)
    Omega = torch.empty((n, 0), dtype=dtype, device=device)
    E = torch.inf
    m = l_0
    break_early = False
    U = torch.Tensor([])
    Lambda_hat = torch.Tensor([])
    nu = math.sqrt(n) * torch.finfo().eps

    while E > error_tol:
        Omega_0 = torch.randn((n, m), dtype=dtype, device=device)
        Omega_0, _ = torch.linalg.qr(Omega_0)
        Y_0 = lo.operator_matrix_product(A, Omega_0)
        Omega = torch.hstack([Omega, Omega_0])
        Y = torch.hstack([Y, Y_0])
        Y_nu = Y + nu * Omega
        C = torch.linalg.cholesky(Omega.T @ Y_nu).T
        B = torch.linalg.solve_triangular(C, Y_nu, upper=True, left=False)
        U, Sigma, _ = torch.linalg.svd(B, False)
        Lambda_hat = torch.relu(Sigma * Sigma - nu)

        if break_early:
            break

        E = randomized_power_err_est(A, U, Lambda_hat, power_iter_count, dtype, device)
        m, l_0 = l_0, 2 * l_0
        if l_0 > l_max:
            l_0 = l_0 - m
            m = l_max - l_0
            break_early = True
    #Add threshold to cut off too small values
    return _NystromApproximation(U, Lambda_hat)


def randomized_power_err_est(A, U, Lambda_hat, q, dtype, device):
    g = torch.randn(A.shape[0], dtype=dtype, device=device)
    v_0 = g / torch.linalg.norm(g)
    E_hat = torch.inf
    for _ in range(q):
        v = A @ v_0 - U @ (Lambda_hat * (U.T @ v_0))
        E_hat = (v_0 * v).sum()
        v = v / torch.linalg.norm(v)
        v_0 = v
    return E_hat



from setuptools import setup

with open("README.md", "r") as fh:
   long_description = fh.read()

setup(
    name="torch-linops",
    version="0.1.3",
    packages=["linops"],
    license="GPLv3",
    description="A library to define abstract linear operators, and associated algebra and matrix-free algorithms, that works with pyTorch Tensors.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    setup_requires=["setuptools>=18.0"],
    install_requires=[
        "torch",
        "scipy"
    ],
    url="https://github.com/cvxgrp/torch_linops",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    author="Parth Nobel",
    author_email="ptnobel@stanford.edu",
)
